use std::sync::mpsc;

use leptos::logging::log;
use protocol::{Error, Signal, SignalType};
use web_sys::{
    ErrorEvent, MessageEvent, WebSocket, js_sys,
    wasm_bindgen::{JsCast, prelude::Closure},
};

static SIGNALLING_ADDRESS: &str = "ws://192.168.1.3:4546";

thread_local! {
static WEBSOCKET: WebSocket = SignallingChannel::init().unwrap();

static OFFER_CHANNEL: (mpsc::Sender<Signal>, mpsc::Receiver<Signal>) = mpsc::channel();
static ANSWER_CHANNEL: (mpsc::Sender<Signal>, mpsc::Receiver<Signal>) = mpsc::channel();
static ICE_CANDIDATE_CHANNEL: (mpsc::Sender<Signal>, mpsc::Receiver<Signal>) = mpsc::channel();
}

struct SignallingChannel {}

impl SignallingChannel {
    fn init() -> Result<WebSocket, Error> {
        let web_socket =
            WebSocket::new(SIGNALLING_ADDRESS).map_err(|_| Error::WebSocketInitialization)?;

        let on_message_callback = Closure::<dyn Fn(_)>::new(move |message_event: MessageEvent| {
            if let Ok(received_data) = message_event.data().dyn_into::<js_sys::JsString>() {
                if let Some(received_data) = received_data.as_string() {
                    if let Ok(received_signal) = serde_json::from_str::<Signal>(&received_data) {
                        if received_signal.get_signal_type() == SignalType::ICECandidate {
                            ICE_CANDIDATE_CHANNEL.with(|ice_candidate_channel| {
                                ice_candidate_channel
                                    .0
                                    .send(received_signal)
                                    .expect("Never");
                            })
                        } else if received_signal.get_signal_type() == SignalType::Offer {
                            OFFER_CHANNEL.with(|offer_channel| {
                                offer_channel.0.send(received_signal).expect("Never");
                            })
                        } else if received_signal.get_signal_type() == SignalType::Answer {
                            ANSWER_CHANNEL.with(|answer_channel| {
                                answer_channel.0.send(received_signal).expect("Never");
                            })
                        }
                    }
                }
            }
        });
        web_socket.set_onmessage(Some(on_message_callback.as_ref().unchecked_ref()));
        on_message_callback.forget();

        let on_error_callback = Closure::<dyn Fn(_)>::new(move |error_event: ErrorEvent| {
            log!("Error: WebSocket | {:#?}", error_event);
        });
        web_socket.set_onerror(Some(on_error_callback.as_ref().unchecked_ref()));
        on_error_callback.forget();

        Ok(web_socket)
    }

    fn send(signal: &Signal) -> Result<(), Error> {
        let json = serde_json::json!(signal);
        WEBSOCKET.with(|websocket| {
            websocket
                .send_with_str(&json.to_string())
                .map_err(|_| Error::WebSocketSend)
        })
    }

    fn receive_offer() -> Result<Signal, Error> {
        OFFER_CHANNEL.with(|offer_channel| {
            offer_channel
                .1
                .recv()
                .map_err(|err_val| Error::OfferChannelReceive(err_val.to_string()))
        })
    }

    fn receive_answer() -> Result<Signal, Error> {
        ANSWER_CHANNEL.with(|answer_channel| {
            answer_channel
                .1
                .recv()
                .map_err(|err_val| Error::AnswerChannelReceive(err_val.to_string()))
        })
    }

    fn receive_ice_candidate() -> Result<Signal, Error> {
        ICE_CANDIDATE_CHANNEL.with(|ice_candidate_channel| {
            ice_candidate_channel
                .1
                .recv()
                .map_err(|err_val| Error::ICECandidateChannelReceive(err_val.to_string()))
        })
    }
}
pub fn send_auth(data: &String) -> Result<(), Error> {
    let offer = Signal::new(&SignalType::Auth, data);

    SignallingChannel::send(&offer)
}

pub fn send_offer(data: &String) -> Result<(), Error> {
    let offer = Signal::new(&SignalType::Offer, data);

    SignallingChannel::send(&offer)
}

pub fn receive_offer() -> Result<Signal, Error> {
    SignallingChannel::receive_offer()
}

pub fn send_answer(data: &String) -> Result<(), Error> {
    let offer = Signal::new(&SignalType::Answer, data);

    SignallingChannel::send(&offer)
}

pub fn receive_answer() -> Result<Signal, Error> {
    SignallingChannel::receive_answer()
}

pub fn send_ice_candidate(data: &String) -> Result<(), Error> {
    let offer = Signal::new(&SignalType::ICECandidate, data);

    SignallingChannel::send(&offer)
}

pub fn receive_ice_candidate() -> Result<Signal, Error> {
    SignallingChannel::receive_ice_candidate()
}
